/*

   RAMDemo1.pde

   This sketch illustrates using the Rugged Circuits QuadRAM shield:
   
                http://ruggedcircuits.com/html/quadram.html
   
   In this sketch, the simplest method of external memory access is demonstrated: no bank 
   control and no high memory masking. Of the 512 kilobytes of RAM on the shield, about
   56 kilobytes is available using this simple method. See the URL above for more details.

   This sketch uses the fixed-point 8-bit FFT function developed for AVR to illustrate
   large-size FFT's. A 16384-point complex FFT is demonstrated. This requires 32768
   bytes of RAM, more than is available on an Arduino Mega, hence the external RAM
   on the QuadRAM shield is used.

   The sketch works as follows:

   * A sinusoid of known frequency is stored in external RAM (set the frequency by
     changing the TestFreq #define statement further down)

   * An FFT is computed on the sinusoid

   * The frequency at which the maximum FFT magnitude occurs is found. This frequency
     should be very close to the TestFreq value (within about Fs/N Hertz where N is
     the FFT size, 16384 in this case, and Fs is the sampling frequency).

   Progress and results are displayed on the serial terminal at 38400bps.

   This software is licensed under the GNU General Public License (GPL) Version
   3 or later. This license is described at
   http://www.gnu.org/licenses/gpl.html
  
   Application Version 1.0 -- October 2011 Rugged Circuits LLC
   http://www.ruggedcircuits.com/html/quadram.html
*/

#include <avr/pgmspace.h>
////////////////////////////////////////////////////////////
//
//     F F T    F U N C T I O N S
//
////////////////////////////////////////////////////////////

/* fix_fft.c - Fixed-point in-place Fast Fourier Transform  */
/*
  All data are fixed-point short integers, in which -128
  to +127 represent -1.0 to +1.0 respectively. Integer
  arithmetic is used for speed, instead of the more natural
  floating-point.

  For the forward FFT (time -> freq), fixed scaling is
  performed to prevent arithmetic overflow, and to map a 0dB
  sine/cosine wave (i.e. amplitude = 128) to two -6dB freq
  coefficients. The return value is always 0.

  Written by:  Tom Roberts  11/8/89
  Made portable:  Malcolm Slaney 12/15/94 malcolm@interval.com
  Enhanced:  Dimitrios P. Bouras  14 Jun 2006 dbouras@ieee.org
  Modified for 8bit values David Keller  10.10.2010
  Modified Rugged Circuits LLC (http://ruggedcircuits.com) 2011.08.27:
    - Replaced pgm_read_word_near with pgm_read_byte_near per forum posting:
         http://www.arduino.cc/cgi-bin/yabb2/YaBB.pl?num=1286718155/17
    - Modified to work with Rugged Circuits MegaRAM shield for Arduino Mega/Mega2560
      (http://ruggedcircuits.com/html/megaram.html)
    - Increased maximum FFT size to 32k (32768)
    - Cleanups, documentation, etc.
    - Removed inverse FFT functionality since it hasn't yet been cleanly ported to 8 bit code
*/


// N_WAVE can't be bigger than 32768 on the AVR
#define N_WAVE    16384U  // full length of Sinewave[]
#define LOG2_N_WAVE 14    // log2(N_WAVE)

/* Since we only use 3/4 of N_WAVE, we define only this many samples, in order
 * to conserve data space.  */

const prog_int8_t Sinewave[N_WAVE-N_WAVE/4] PROGMEM = {
/* Sinewave table with 2**14 entries (less the last quarter of the table). A
   table can be generated by the Python program:

       from math import *
       N=16384     # Table size, must be a power of 2
       for k in range(N - N/4):
         print int(127.5*sin(2*pi*k/N)), ","
*/
#include "sinewave14.h"   
};

/* FIX_MPY() - fixed-point multiplication & scaling.
   Scaling ensures that result remains 8-bit.  */

inline int8_t FIX_MPY(int8_t a, int8_t b)
{
  /* Two 8-bit signed numbers ([-128,127]) multiplied together will lead to
   * a result in the range [-16256,16384]. This is a 15-bit result. We therefore
   * need to shift the result to the right by 7 bits to get back to an 8-bit range. */

    /* Multiply, then shift right by only 6 bits (for now) prior to rounding */
    int16_t c = ((int16_t)a * (int16_t)b) >> 6;

    /* last bit shifted out = rounding-bit */
    b = c & 0x01;
    /* last shift + rounding bit */
    a = (c >> 1) + b;

    return a;
}

/* fix_fft() - perform forward fast Fourier transform.
  fr[n],fi[n] are real and imaginary arrays, both INPUT AND RESULT (in-place
  FFT), with 0 <= n < 2**m

  Currently must have m<=15.

  Return value is -1 on error (basically, m is too big) or 0 for success.  */

int8_t fix_fft(int8_t fr[], int8_t fi[], uint8_t log2N)
{
    int8_t qr, qi, tr, ti, wr, wi;
    uint16_t n, l, mr, istep, i, j, m;
    int8_t k;

    // Compute the length of the FFT: n=2**m
    n = (uint16_t)1 << log2N;

    /* max FFT size = N_WAVE */
    if (n > N_WAVE) return -1;

    mr = 0;

    /* decimation in time - re-order data */
    for (m=1; m<n; ++m) {
      l = n;
      do {
        l >>= 1;
      } while (mr+l >= n);
      mr = (mr & (l-1)) + l;

      if (mr <= m) continue;

      // Swap fr[m] with fr[mr]
      tr = fr[m];
      fr[m] = fr[mr];
      fr[mr] = tr;

      // Swap fi[m] with fi[mr]
      ti = fi[m];
      fi[m] = fi[mr];
      fi[mr] = ti;
    }

    l = 1;
    k = LOG2_N_WAVE-1;
    while (l < n) {
      /* fixed scaling, for proper normalization --
        there will be log2(n) passes, so this results in an overall factor of
        1/n, distributed to maximize arithmetic accuracy.  It may not be
        obvious, but the shift will be performed on each data point exactly
        once, during this pass.  */ 
      istep = l << 1;
      for (m=0; m<l; ++m) {
        j = m << k;

        /* At this point, 0 <= j < N_WAVE/2 */
        wr =  (int8_t)pgm_read_byte_near(Sinewave + j + N_WAVE/4) >> 1;
        wi = (-(int8_t)pgm_read_byte_near(Sinewave + j)) >> 1;

        for (i=m; i<n; i += istep) {
            j = i + l;
            tr = FIX_MPY(wr,fr[j]) - FIX_MPY(wi,fi[j]);
            ti = FIX_MPY(wr,fi[j]) + FIX_MPY(wi,fr[j]);
            qr = fr[i] >> 1;
            qi = fi[i] >> 1;

            fr[j] = qr - tr;
            fi[j] = qi - ti;
            fr[i] = qr + tr;
            fi[i] = qi + ti;
        }
      }
      --k;
      l = istep;
    }
    return 0; // success
}

////////////////////////////////////////////////////////////
//
//     A R D U I N O    F U N C T I O N S
//
////////////////////////////////////////////////////////////

#define Fs       20000 // Sampling frequency (arbitrary for testing purposes)
#define TestFreq 820   // Must be in the range 0 <= TestFreq < Fs/2

// These will be pointers into MegaRAM shield external RAM, offset by N_WAVE bytes
int8_t *real;
int8_t *imag;

void setup(void)
{
  uint16_t i, maxi;
  double maxmag;

  Serial.begin(38400);

#if N_WAVE > 16384
#  error This code will only work for N_WAVE <= 16384
#endif

  // Enable MegaRAM shield external memory with no wait states
  XMCRA = _BV(SRE);

  // Disable high memory masking (don't need >56kbytes in this app)
  XMCRB = 0;

  // Enable RAM device (PD7 or D38, active low)
  pinMode(38, OUTPUT); digitalWrite(38, LOW);

  // Enable bank select outputs and select bank 0. Don't need multiple banks in this app.
  pinMode(42, OUTPUT); digitalWrite(42, LOW);   // PL7 or D42
  pinMode(43, OUTPUT); digitalWrite(43, LOW);   // PL6 or D43
  pinMode(44, OUTPUT); digitalWrite(44, LOW);   // PL5 or D44

  // First fill up an N_WAVE array with a sinusoid of known frequency
  Serial.println("Filling RAM with sinusoid ...");
  real = (int8_t *) 0x8000U; // External RAM on MegaRAM shield
  imag = real + N_WAVE;

  for (i = 0; i < N_WAVE; i++) {
    real[i] = (int8_t) (63*sin(2*M_PI*TestFreq/Fs*i));
    imag[i] = 0;
  }

  // Now compute the Fourier transform
  Serial.println("Computing FFT ...");
  (void) fix_fft(real, imag, LOG2_N_WAVE);

  // Find the index where the largest magnitude component exists
  Serial.println("Finding largest component ...");
  for (i=0, maxi=0, maxmag=0; i < N_WAVE/2; i++) {
    double mag;

    // FFT result is complex-valued, so compute the squared magnitude to get a measure
    // of relative power.
    mag = double(real[i])*real[i] + double(imag[i])*imag[i];
    if (mag > maxmag) {
      maxmag = mag;
      maxi = i;
    }
  }

  // The frequency of the highest component is given by this expression
  double maxf = double(maxi)/N_WAVE*Fs;

  Serial.print("FFT says strongest frequency component is at ");
  Serial.print(maxf);
  Serial.println(" Hz");
}

void loop(void)
{
}

// vim: syntax=cpp ai cindent sw=2 ts=2 expandtab
